// STLprofiler.cpp : Definiert den Einsprungpunkt fr die Konsolenanwendung.
//

#ifdef USE_STLPORT
# include "vector"
# include "deque"
# include "list"
# include "iostream"
# include "iomanip"
#else
# include <vector>
# include <deque>
# include <list>
# include <iostream>
# include <iomanip>
#endif

#include <ctime>

//////////////////////////////////////////////////////////////////////////////
//
// test configuration
//

template<class ContainerT>
struct config
{
  config(ContainerT& testee, int nElements, size_t nIterations)
    : testee(testee), nElements(nElements), nIterations(nIterations)
  {
  }

  ContainerT& testee;
  const size_t nElements;
  const int nIterations;
};

//////////////////////////////////////////////////////////////////////////////
//
// runtime measurement
//

template<class ContainerT>
double
measure_runtime(
    config<ContainerT>& cnf,
    void (*prepare)(config<ContainerT>&),
    bool (*test)(config<ContainerT>&))
{
  bool valid_test = false;
  clock_t total_clocks = 0;
  for (int i = 1; i <= cnf.nIterations; ++i)
  {
    prepare(cnf);
    clock_t begin = ::clock();
    valid_test = test(cnf);
    clock_t end = ::clock();
    
    total_clocks += (end - begin);
  }
  if (!valid_test)
  {
    // no sensible runtime for invalid test

    total_clocks = -CLOCKS_PER_SEC;
  }
  return static_cast<double>(total_clocks) / CLOCKS_PER_SEC;
}


//////////////////////////////////////////////////////////////////////////////
//
// test preparation
//

template<class ContainerT>
struct plain_preparation
{
  static void fill(config<ContainerT>& cnf)
  {
    ContainerT ().swap(cnf.testee);
  }
  
  static void traverse(config<ContainerT>& cnf)
  {
  }
  
  static void access(config<ContainerT>& cnf)
  {
  }
};


template<class ContainerT>
struct reserving_preparation
  : plain_preparation<ContainerT>
{
  static void fill(config<ContainerT>& cnf)
  {
    plain_preparation<ContainerT>::fill(cnf);
    cnf.testee.reserve(cnf.nElements);
  }
};


//////////////////////////////////////////////////////////////////////////////
//
// test operations
//

template<class ContainerT>
struct rc_test
{
  static bool fill(config<ContainerT>& cnf)
  {
    // perform test

    for (size_t i = 0; i < cnf.nElements; ++i)
    {
      cnf.testee.push_back(i);
    }
    return true;
  }
  
  static bool traverse(config<ContainerT>& cnf)
  {
    // traversing elements
    
    ContainerT::const_iterator container_walker = cnf.testee.begin();
    ContainerT::const_iterator walk_terminator = cnf.testee.end();
    while (container_walker != walk_terminator)
    {
      ++container_walker;
      int element = *container_walker;
      element;
    }
    return true;
  }
  
  static bool access(config<ContainerT>& cnf)
  {
    // no random access supported

    return false;
  }
};

template<class ContainerT>
struct rac_test
  : public rc_test<ContainerT>
{
  static bool access(config<ContainerT>& cnf)
  {
    // accessing elements
  
    const size_t n_elements = cnf.testee.size();
    for (size_t i = 0; i < n_elements; ++i)
    {
      int element = cnf.testee[i];
      element;
    }
    return true;
  }
};


//////////////////////////////////////////////////////////////////////////////
//
// test execution
//

void print_value(std::ostream & target, double value, size_t width)
{
  using std::setw;
  using std::setprecision;

  target << "\t" << setw(width);
  if (-1 == value)
  {
    target << "N/A";
  }
  else
  {
    target << setprecision(3) << value;
  }
}

void
print_results(std::ostream & target, double cr, double it, double ac)
{
  using std::endl;
  using std::setw;
  
  const CELL_WIDTH = 10;

  target
      << "\t" << setw(CELL_WIDTH) << "creation"
      << "\t" << setw(CELL_WIDTH) << "iteration"
      << "\t" << setw(CELL_WIDTH) << "access"
      << endl;
  char old_fill = target.fill('-');
  target
      << "\t" << setw(CELL_WIDTH) << ""
      << "\t" << setw(CELL_WIDTH) << ""
      << "\t" << setw(CELL_WIDTH) << ""
      << endl;
  target.fill(old_fill);
  
  print_value(target, cr, CELL_WIDTH);
  print_value(target, it, CELL_WIDTH);
  print_value(target, ac, CELL_WIDTH);
  target << endl;
}

template<class ContainerT, class PrepT, class TestT>
void
perform_test(ContainerT& testee, size_t nElements, size_t nIterations, const PrepT* prep, const TestT* test)
{
  config<ContainerT> cnf(testee, nElements, nIterations);

  double cr_time =
      measure_runtime(cnf, PrepT::fill, TestT::fill);
  
  double it_time =
      measure_runtime(cnf, PrepT::traverse, TestT::traverse);
  
  double ac_time =
      measure_runtime(cnf, PrepT::access, TestT::access);
  
  // freeing memory
  
  ContainerT ().swap(testee);
  
  print_results(std::cout, cr_time, it_time, ac_time);
}


//////////////////////////////////////////////////////////////////////////////
//
// concrete test-suites
//

// random-access container, no reserve
template<class ContainerT>
plain_rac_test(ContainerT& testee, size_t nElements, size_t nIterations)
{
  typedef plain_preparation<ContainerT> PrepT;
  typedef rac_test<ContainerT> TestT;
  perform_test(testee, nElements, nIterations,
      static_cast<const PrepT*>(0), static_cast<const TestT*>(0));
}

// random-access container, reserving
template<class ContainerT>
reserving_rac_test(ContainerT& testee, size_t nElements, size_t nIterations)
{
  typedef reserving_preparation<ContainerT> PrepT;
  typedef rac_test<ContainerT> TestT;
  perform_test(testee, nElements, nIterations,
      static_cast<const PrepT*>(0), static_cast<const TestT*>(0));
}

// reversible container, no reserve
template<class ContainerT>
plain_rc_test(ContainerT& testee, size_t nElements, size_t nIterations)
{
  typedef plain_preparation<ContainerT> PrepT;
  typedef rc_test<ContainerT> TestT;
  perform_test(testee, nElements, nIterations,
      static_cast<const PrepT*>(0), static_cast<const TestT*>(0));
}

//////////////////////////////////////////////////////////////////////////////
//
// entry point
//

int
main(int argc, char* argv[])
{
  using std::cout;
  using std::endl;
  using std::hex;

  cout << "Profiling "
#ifdef USE_STLPORT
      << "STLport "
      << hex << _STLPORT_VERSION / (16 * 16) << "."
      << hex << _STLPORT_VERSION % (16 * 16)
#else
      << "MSVC 6.0"
#endif
      << " STL implementation." << endl << endl;

  // configuration area
  
  const N_ELEMENTS = 10000000;
  const N_ITERATION = 5;
  typedef int ELEMENT_TYPE;
  
  cout << endl << "testing std::vector (no reserve)" << endl << endl;
  std::vector<ELEMENT_TYPE> test_vector;
  plain_rac_test(test_vector, N_ELEMENTS, N_ITERATION);

  cout << endl << "testing std::vector (reserving)" << endl << endl;
  std::vector<ELEMENT_TYPE> reserved_test_vector;
  reserving_rac_test(reserved_test_vector, N_ELEMENTS, N_ITERATION);
  
  cout << endl << "testing std::deque" << endl << endl;
  std::deque<ELEMENT_TYPE> test_deque;
  plain_rac_test(test_deque, N_ELEMENTS, N_ITERATION);


  cout << endl << "testing std::list" << endl << endl;
  std::list<int> test_list;
  plain_rc_test(test_list, N_ELEMENTS, N_ITERATION);

  
  return 0;
}
